#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "rapidjson/document.h"
#include "NcClosers.h"
#include "Configuration.h"
#include "Constants.h"
#include "Messages.h"
#include "NcPrintErr.h"
#include "SendTask.h"
#include "NcSocket.h"
#include "NcUart.h"

SendState g_send;

using namespace rapidjson;

bool ConfigureUart(tSendConfig * c)
{
	return UartConfigure(g_send.fd, c->getSpeed(), UART_READ_TIMEOUT_MS);
}

bool OpenSocket(tSendConfig * c)
{
	return SocketOpenServer(c->getSendPort(), &g_send.socketFd, &g_send.address);
}

bool OpenUart(tSendConfig * c)
{
	return UartOpen(c->getUart(), &g_send.fd);
}

bool SendTaskInitInternal(Configuration * c)
{
	memset(&g_send, 0, sizeof(g_send));
	g_send.config = c;
	if (!OpenUart(c) || !OpenSocket(c))
		return false;
	g_send.state = Started;
	return true;
}

bool SendTaskInit(void * p)
{
	Configuration * c = (Configuration *)p;
	return SendTaskInitInternal(c);
}

bool SendTaskShutdown()
{
	g_send.state = Stopped;
	return true;
}

void SendTaskCloseUart()
{
	FdClose(&g_send.fd);
}

void SendTaskCloseSockets()
{
	FdClose(&g_send.socketFd);
}

void SendTaskClose()
{
	SendTaskCloseUart();
	SendTaskCloseSockets();
}

bool ReadClientMsg()
{
	int bufSize = SOCKET_INITIAL_BUFFER;
	uint8_t * p = (uint8_t*)malloc(bufSize);
	if (p==NULL)
	{
		puts("ERROR: out of memory, could not allocate socket buffer");
		return false;
	}
	int readSoFar = 0;
	while (true)
	{
		int n = read(g_send.clientFd, p+readSoFar, bufSize-readSoFar);
		bool finished = (n==0) || (n<0 && errno==EAGAIN) || (n<0 && errno==EWOULDBLOCK);
		if (finished)
			break;
		if (n<0 && !finished)
		{
			PrintErr("Failed to read from socket. errno=%s", PrintErrNo(errno));
			puts(g_error);
			free((void*)p);
			return false;
		}
		else
		{
			readSoFar += n;
			if (bufSize-readSoFar < 64)
			{
				p = (uint8_t *)realloc(p, bufSize + SOCKET_INITIAL_BUFFER);
				if (p==NULL)
				{
					puts("ERROR: out of memory, could not reallocate socket buffer");
					return false;
				}
			}
		}
	}
	g_send.clientRaw = p;
	g_send.clientRawLen = readSoFar;
	return true;
}

void CleanUpClient()
{
	FdClose(&g_send.clientFd);
	Free((void**)&g_send.clientRaw);
	Free((void**)&g_send.clientMsg);
	Free((void**)&g_send.ncSendParams.payload);
}

void WriteStatusToClient(const char * content)
{
	ReplyMsg rm =
		{
			content==NULL || strlen(content)==0 ? "ok" : "error",
			content
		};
	const char * ser = g_send.config->getSerializer()->Serialize(&rm);
	printf("-->%s\n", ser);
	const char * del = g_send.config->getDelimiterFormatted();
	int delLen = strlen(del);
	int len = strlen(ser);
	int finalLen = len + delLen;
	char * final = (char *)malloc(finalLen);
	memcpy(final, ser, len);
	memcpy(final+len, del, delLen);
	// dummy, the posix write is declared with __WUR
	ssize_t dummy = send(g_send.clientFd, final, finalLen, MSG_NOSIGNAL);
	(void) dummy;

	free((void*)ser);
	free((void*)final);
}

bool WaitForMessageToBeWritten()
{
	const int timeOutSecs = g_send.config->getCtsTimeoutSecs();
	struct timespec tvend;
	struct timespec tv;
	clock_gettime(CLOCK_MONOTONIC_RAW, &tvend);
	tvend.tv_sec += timeOutSecs;
	while(!g_send.msgWasWritten)
	{
		usleep(SOCKET_LOOP_PAUSE_MS);
		clock_gettime(CLOCK_MONOTONIC_RAW, &tv);
		if (tv.tv_sec>tvend.tv_sec)
			return false;
	}
	return true;
}

const char* SendMessageToNcApi()
{
	g_send.msgWasWritten = false;
	NcApiErrorCodes ncRet=NCAPI_OK;
	switch (g_send.ncApiType)
	{
		case CommandAcknowledgedEnum:
			ncRet = NcApiSendAcknowledged(0, &g_send.ncSendParams);
			break;
		case WesCmdEnum:
			ncRet = NcApiSendWesCmd(0, &g_send.ncWesCmd);
			break;
		case AltCmdEnum:
			ncRet = NcApiSendAltCmd(0, &g_send.ncAltCmd);
			break;
		case WesResponseEnum:
			ncRet = NcApiSendWesResponse(0, &g_send.ncWesResponse);
			break;
		case CommandRawEnum:
			ncRet = NcApiSendRaw(0, &g_send.ncSendParams);
			break;
			
		default:
			break;
	}
	if (ncRet!=NCAPI_OK)
	{
		PrintErr("Staging data failed: %d", ncRet);
		return g_error;
	}
	if (!WaitForMessageToBeWritten())
	{
		NcApiCancelEnqueuedMessage(0);
		const int timeOutSecs = g_send.config->getCtsTimeoutSecs();
		PrintErr("Message was not written to UART within %d seconds", timeOutSecs);
		return g_error;
	}
	return NULL;
}

void PrepareNcApiFromIncoming(const SendMsg & sm, tNcApiSendParams & m)
{
	m.callbackToken = NULL;
	m.destNodeId = sm.destNodeId;
	m.destPort = sm.destPort;
	m.payload = sm.payload;
	m.payloadLength = sm.payloadLength;
}

void PrintMsg()
{
	uint8_t * p = g_send.clientRaw;
	int len = g_send.clientRawLen;
	uint16_t i;
	p[len] =  0;
	printf("Raw socket data received: ");
	for (i=0; i<len; i++)
		printf( "%02x ", p[i] );
	printf("\r\n<--%s\n", p);

}
const char* ParseMsgSendParams()
{
	memset(&g_send.ncSendParams, 0, sizeof(g_send.ncSendParams));
	SendMsg m;
	const char * ret = g_send.config->getSerializer()->DeSerialize((const char*)g_send.clientRaw, g_send.clientRawLen, &m);
	
	if (m.msgType==CommandRawEnum)
	{
		// Raw
		g_send.ncApiType = CommandRawEnum;
		
	}
	else
	{
		if (ret!=NULL) return ret;
		g_send.ncApiType = CommandAcknowledgedEnum;
		if (m.msgType!=CommandAcknowledgedEnum)
			return "Invalid msgType";
	}
	PrepareNcApiFromIncoming(m, g_send.ncSendParams);
	return NULL;
}

const char* ParseMsgWesCmd()
{
	memset(&g_send.ncWesCmd, 0, sizeof(g_send.ncWesCmd));
	tNcApiWesCmd m;
	const char * ret = g_send.config->getSerializer()->DeSerialize((const char*)g_send.clientRaw, g_send.clientRawLen, &m);
	if (ret!=NULL) return ret;
	g_send.ncApiType = WesCmdEnum;
	g_send.ncWesCmd.cmd = m.cmd;
	return ret;
}

const char* ParseMsgWesResponse()
{
	memset(&g_send.ncWesResponse, 0, sizeof(g_send.ncWesResponse));
	tNcApiWesResponse m;
	const char * ret = g_send.config->getSerializer()->DeSerialize((const char*)g_send.clientRaw, g_send.clientRawLen, &m);
	if (ret!=NULL) return ret;
	g_send.ncApiType = WesResponseEnum;
	g_send.ncWesResponse.nodeId = m.nodeId;
	memcpy( g_send.ncWesResponse.uid, m.uid, 5 );
	return ret;
}

const char* ParseMsg()
{
	const char * ret = ParseMsgSendParams();
	if (ret!=NULL ) ret = ParseMsgWesCmd();
	if (ret!=NULL ) ret = ParseMsgWesResponse();
	return ret;
}

void SendTaskHandleClient(int fd)
{
	g_send.clientFd = fd;
	if (!ReadClientMsg())
		return;
	PrintMsg();
	const char * content = ParseMsg();
	if (content==NULL)
		content = SendMessageToNcApi();
	WriteStatusToClient(content);
	CleanUpClient();
}

void SendTaskAcceptSockets()
{
	int clientFd;
	while (SocketAcceptNonBlock(g_send.socketFd, &clientFd))
		SendTaskHandleClient(clientFd);
}

void SendTaskSleep()
{
	usleep(SOCKET_LOOP_PAUSE_MS * 1000);
}

void SendTaskLoop()
{
	SendTaskAcceptSockets();
	SendTaskSleep();
}

void* SendTaskRun(void * p)
{
	tSendConfig * c = (tSendConfig *)p;
    if (!ConfigureUart(c))
    	return NULL;
	g_send.state = Started;
	while(true)
	{
		switch (g_send.state)
		{
		case Stopped:
			SendTaskClose();
			return NULL;
		case Started:
			SendTaskLoop();
			break;
		case ReLoadConfig:
			SendTaskClose();
			g_send.state = SendTaskInit(g_send.config)
					? Started
					: Stopped;
			break;
		}
	}
	return NULL;
}
